# Copyright (c) 2016 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from tempest.test_discover import plugins


class NetworkingMidonetTempestPlugin(plugins.TempestPlugin):
    def get_opt_lists(self):
        return []

    def load_tests(self):
        this_dir = os.path.dirname(os.path.abspath(__file__))
        # top_level_dir = $(this_dir)/../../../..
        d = os.path.split(this_dir)[0]
        d = os.path.split(d)[0]
        d = os.path.split(d)[0]
        top_level_dir = os.path.split(d)[0]
        test_dir = os.path.join(top_level_dir,
                                'midonet/neutron/tests/tempest/tests')
        return (test_dir, top_level_dir)

    def register_opts(self, conf):
        return



# Copyright (C) 2015 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import constants as const
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import excutils

from neutron_vpnaas.services.vpn import plugin
from neutron_vpnaas.services.vpn.service_drivers import base_ipsec
from neutron_vpnaas.services.vpn.service_drivers import ipsec_validator

from midonet.neutron.client import base as c_base
from midonet.neutron.common import config  # noqa

LOG = logging.getLogger(__name__)


# TODO(yamamoto): Introduce VPNaaS PRECOMMIT callbacks and
# subscribe them for task-based api.
class MidonetIPsecVPNDriver(base_ipsec.BaseIPsecVPNDriver):
    def __init__(self, service_plugin):
        super(MidonetIPsecVPNDriver, self).__init__(
            service_plugin, ipsec_validator.IpsecVpnValidator(self))
        self.plugin = plugin.VPNPlugin()
        self.client = c_base.load_client(cfg.CONF.MIDONET)

    def create_rpc_conn(self):
        pass

    def create_vpnservice(self, context, vpnservice_dict):
        super(MidonetIPsecVPNDriver, self).create_vpnservice(
            context, vpnservice_dict)
        try:
            self.client.create_vpn_service(context, vpnservice_dict)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to create a vpn_service %(service_id)s "
                          "in MidoNet: %(err)s",
                          {"service_id": vpnservice_dict["id"], "err": ex})
                try:
                    self.plugin.delete_vpnservice(
                        context, vpnservice_dict['id'])
                except Exception:
                    LOG.exception("Failed to delete vpn_service %s",
                                  vpnservice_dict['id'])

        self.update_vpn_service_status(context, vpnservice_dict['id'],
                                       const.ACTIVE)

    def update_vpnservice(self, context, old_vpnservice, vpnservice):
        try:
            self.client.update_vpn_service(
                context, vpnservice['id'], vpnservice)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to update a vpn_service %(service_id)s "
                          "in MidoNet: %(err)s",
                          {"service_id": vpnservice["id"], "err": ex})
                try:
                    self.update_vpn_service_status(
                        context, vpnservice['id'], const.ERROR)
                except Exception:
                    LOG.exception("Failed to update vpn_service status %s",
                                  vpnservice['id'])

    def delete_vpnservice(self, context, vpnservice):
        try:
            self.client.delete_vpn_service(context, vpnservice['id'])
        except Exception:
            LOG.exception("Failed to delete vpn_service %s",
                          vpnservice['id'])

    def create_ipsec_site_connection(self, context, ipsec_site_connection):
        ipsec_site_conn_info = self.make_ipsec_site_connection_dict(
            context, ipsec_site_connection['id'])
        try:
            self.client.create_ipsec_site_conn(context, ipsec_site_conn_info)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to create a ipsec_site_connection "
                          "%(conn_id)s in MidoNet: %(err)s",
                          {"conn_id": ipsec_site_connection['id'], "err": ex})
                try:
                    self.plugin.delete_ipsec_site_connection(
                        context, ipsec_site_connection['id'])
                except Exception:
                    LOG.exception("Failed to delete ipsec_site_connection %s",
                                  ipsec_site_connection['id'])

        self.service_plugin.update_ipsec_site_conn_status(
            context, ipsec_site_connection['id'], const.ACTIVE)

    def update_ipsec_site_connection(self, context, old_ipsec_site_connection,
                                     ipsec_site_connection):
        ipsec_site_conn_info = self.make_ipsec_site_connection_dict(
            context, ipsec_site_connection['id'])
        try:
            self.client.update_ipsec_site_conn(
                context, ipsec_site_connection['id'], ipsec_site_conn_info)
        except Exception as ex:
            with excutils.save_and_reraise_exception():
                LOG.error("Failed to update a ipsec_site_connection "
                          "%(service_id)s in MidoNet: %(err)s",
                          {"service_id": ipsec_site_connection['id'],
                           "err": ex})
                try:
                    self.service_plugin.update_ipsec_site_conn_status(
                        context, ipsec_site_connection['id'], const.ERROR)
                except Exception:
                    LOG.exception("Failed to update ipsec_site_connection "
                                  "status %s",
                                  ipsec_site_connection['id'])

    def delete_ipsec_site_connection(self, context, ipsec_site_connection):
        try:
            self.client.delete_ipsec_site_conn(
                context, ipsec_site_connection['id'])
        except Exception:
            LOG.error("Failed to delete ipsec_site_connection %s",
                      ipsec_site_connection['id'])

    def make_ipsec_site_connection_dict(self, context, ipsec_site_conn_id):
        ipsec_site_conn = self.service_plugin._get_ipsec_site_connection(
            context, ipsec_site_conn_id)
        vpnservice = ipsec_site_conn.vpnservice

        local_cidr_map = self.service_plugin._build_local_subnet_cidr_map(
            context)
        vpnservice_dict = self.make_vpnservice_dict(vpnservice, local_cidr_map)
        ipsec_site_conn_dict = list(filter(
            lambda conn: conn['id'] == ipsec_site_conn_id,
            vpnservice_dict['ipsec_site_connections']))[0]
        del ipsec_site_conn_dict['vpnservice']

        return ipsec_site_conn_dict

    def update_vpn_service_status(self, context, vpnservice_id, status):
        # this method is used only for updating a vpn_service status
        self.service_plugin.update_status_by_agent(
            context,
            [{'id': vpnservice_id,
              'status': status,
              'updated_pending_status': True,
              'ipsec_site_connections': {}
              }])



# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import neutron.db.api as db

from midonet.neutron.client import base
from midonet.neutron.db import task_db as task


class MidonetClusterClient(base.MidonetClientBase):

    def __init__(self, conf):
        self.conf = conf

    def initialize(self):
        task.create_config_task(db.get_writer_session(), dict(self.conf))

    def create_network_precommit(self, context, network):
        task.create_task(context, task.CREATE, data_type=task.NETWORK,
                         resource_id=network['id'], data=network)

    def update_network_precommit(self, context, network_id, network):
        task.create_task(context, task.UPDATE, data_type=task.NETWORK,
                         resource_id=network_id, data=network)

    def delete_network_precommit(self, context, network_id):
        task.create_task(context, task.DELETE, data_type=task.NETWORK,
                         resource_id=network_id)

    def create_subnet_precommit(self, context, subnet):
        task.create_task(context, task.CREATE, data_type=task.SUBNET,
                         resource_id=subnet['id'], data=subnet)

    def update_subnet_precommit(self, context, subnet_id, subnet):
        task.create_task(context, task.UPDATE, data_type=task.SUBNET,
                         resource_id=subnet_id, data=subnet)

    def delete_subnet_precommit(self, context, subnet_id):
        task.create_task(context, task.DELETE, data_type=task.SUBNET,
                         resource_id=subnet_id)

    def create_port_precommit(self, context, port):
        task.create_task(context, task.CREATE, data_type=task.PORT,
                         resource_id=port['id'], data=port)

    def update_port_precommit(self, context, port_id, port):
        task.create_task(context, task.UPDATE, data_type=task.PORT,
                         resource_id=port_id, data=port)

    def delete_port_precommit(self, context, port_id):
        task.create_task(context, task.DELETE, data_type=task.PORT,
                         resource_id=port_id)

    def create_router_precommit(self, context, router):
        task.create_task(context, task.CREATE, data_type=task.ROUTER,
                         resource_id=router['id'], data=router)

    def update_router_precommit(self, context, router_id, router):
        task.create_task(context, task.UPDATE, data_type=task.ROUTER,
                         resource_id=router_id, data=router)

    def delete_router_precommit(self, context, router_id):
        task.create_task(context, task.DELETE, data_type=task.ROUTER,
                         resource_id=router_id)

    def create_floatingip_precommit(self, context, floatingip):
        task.create_task(context, task.CREATE, data_type=task.FLOATING_IP,
                         resource_id=floatingip['id'], data=floatingip)

    def update_floatingip_precommit(self, context, floatingip_id, floatingip):
        task.create_task(context, task.UPDATE, data_type=task.FLOATING_IP,
                         resource_id=floatingip_id, data=floatingip)

    def delete_floatingip_precommit(self, context, floatingip_id):
        task.create_task(context, task.DELETE, data_type=task.FLOATING_IP,
                         resource_id=floatingip_id)

    def create_security_group_precommit(self, context, security_group):
        task.create_task(context, task.CREATE, data_type=task.SECURITY_GROUP,
                         resource_id=security_group['id'], data=security_group)

    def delete_security_group_precommit(self, context, security_group_id):
        task.create_task(context, task.DELETE, data_type=task.SECURITY_GROUP,
                         resource_id=security_group_id)

    def create_security_group_rule_precommit(self, context,
                                             security_group_rule):
        task.create_task(context, task.CREATE,
                         data_type=task.SECURITY_GROUP_RULE,
                         resource_id=security_group_rule['id'],
                         data=security_group_rule)

    def delete_security_group_rule_precommit(self, context,
                                             security_group_rule_id):
        task.create_task(context, task.DELETE,
                         data_type=task.SECURITY_GROUP_RULE,
                         resource_id=security_group_rule_id)

    # LBaaS

    def create_vip(self, context, vip):
        task.create_task(context, task.CREATE, data_type=task.VIP,
                         resource_id=vip['id'], data=vip)

    def update_vip(self, context, vip_id, vip):
        task.create_task(context, task.UPDATE, data_type=task.VIP,
                         resource_id=vip_id, data=vip)

    def delete_vip(self, context, vip_id):
        task.create_task(context, task.DELETE, data_type=task.VIP,
                         resource_id=vip_id)

    def create_pool(self, context, pool):
        task.create_task(context, task.CREATE, data_type=task.POOL,
                         resource_id=pool['id'], data=pool)

    def update_pool(self, context, pool_id, pool):
        task.create_task(context, task.UPDATE, data_type=task.POOL,
                         resource_id=pool_id, data=pool)

    def delete_pool(self, context, pool_id):
        task.create_task(context, task.DELETE, data_type=task.POOL,
                         resource_id=pool_id)

    def create_member(self, context, member):
        task.create_task(context, task.CREATE, data_type=task.MEMBER,
                         resource_id=member['id'], data=member)

    def update_member(self, context, member_id, member):
        task.create_task(context, task.UPDATE, data_type=task.MEMBER,
                         resource_id=member_id, data=member)

    def delete_member(self, context, member_id):
        task.create_task(context, task.DELETE, data_type=task.MEMBER,
                         resource_id=member_id)

    def create_health_monitor(self, context, health_monitor):
        task.create_task(context, task.CREATE, data_type=task.HEALTH_MONITOR,
                         resource_id=health_monitor['id'], data=health_monitor)

    def update_health_monitor(self, context, health_monitor_id,
                              health_monitor):
        task.create_task(context, task.UPDATE, data_type=task.HEALTH_MONITOR,
                         resource_id=health_monitor_id, data=health_monitor)

    def delete_health_monitor(self, context, health_monitor_id):
        task.create_task(context, task.DELETE, data_type=task.HEALTH_MONITOR,
                         resource_id=health_monitor_id)

    # TODO(yamamoto): Implement the following
    # - FWaaS v1
    # - VPNaaS
    # - L2 gateway
    # - gateway device
    # - BGP
    # - logging resource / firewall log
    # - Tap as a service
    # - QoS
    # - LBaaS v2



# Copyright (C) 2016 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutronclient.common import extension
from neutronclient.neutron import v2_0 as loggingV20

from midonet.neutron._i18n import _


def add_common_paramaters_to_arguments(parser):
    parser.add_argument(
        '--description', dest='description',
        help=_('Description of firewall log.'))
    parser.add_argument(
        '--fw-event', dest='fw_event',
        metavar='<ALL | ACCEPT | DROP>',
        choices=['ALL', 'ACCEPT', 'DROP'],
        help=_('Event of firewall. Defaults to ALL.'))


def _get_logging_resource_id(client, logging_resource_id_or_name):
    return loggingV20.find_resourceid_by_name_or_id(
        client, 'logging_resource',
        logging_resource_id_or_name)


class FirewallLog(extension.NeutronClientExtension):
    parent_resource = 'logging_resources'
    resource = 'firewall_log'
    resource_plural = 'firewall_logs'
    object_path = '/logging/%s/%%s/%s' % (parent_resource, resource_plural)
    resource_path = '/logging/%s/%%s/%s/%%%%s' % (parent_resource,
                                                  resource_plural)
    versions = ['2.0']

    def add_known_arguments(self, parser):
        super(FirewallLog, self).add_known_arguments(parser)
        parser.add_argument(
            'logging_resource', metavar='LOGGING_RESOURCE',
            help=_('ID of the logging_resource.'))
        add_common_paramaters_to_arguments(parser)

    def set_extra_attrs(self, parsed_args):
        self.parent_id = _get_logging_resource_id(self.get_client(),
                                                  parsed_args.logging_resource)


class FirewallLogCreate(extension.ClientExtensionCreate, FirewallLog):
    """Create Logging Resource Firewall Log information."""

    shell_command = 'logging-firewall-create'

    def add_known_arguments(self, parser):
        super(FirewallLogCreate, self).add_known_arguments(parser)
        parser.add_argument(
            '--firewall-id', dest='firewall_id',
            required=True,
            help=_('Firewall to be logged'))

    def args2body(self, args):
        body = {}
        attributes = ['firewall_id', 'description', 'fw_event']
        loggingV20.update_dict(args, body, attributes)
        return {'firewall_log': body}


class FirewallLogList(extension.ClientExtensionList, FirewallLog):
    """List Logging Resource Firewall Logs."""

    shell_command = 'logging-firewall-list'
    list_columns = ['id', 'firewall_id', 'fw_event']
    pagination_support = True
    sorting_support = True


class FirewallLogShow(extension.ClientExtensionShow, FirewallLog):
    """Show information of a given Logging Resource Firewall Log."""

    shell_command = 'logging-firewall-show'
    allow_names = False


class FirewallLogDelete(extension.ClientExtensionDelete, FirewallLog):
    """Delete a given Logging Resource Firewall Log."""

    shell_command = 'logging-firewall-delete'
    allow_names = False


class FirewallLogUpdate(extension.ClientExtensionUpdate, FirewallLog):
    """Update a given Logging Resource Firewall Log."""

    shell_command = 'logging-firewall-update'

    def args2body(self, args):
        body = {}
        attributes = ['description', 'fw_event']
        loggingV20.update_dict(args, body, attributes)

        return {'firewall_log': body}



# Copyright (C) 2017 Midokura SARL.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import functools

from oslo_log import log as logging
from oslo_serialization import jsonutils
from oslo_utils import uuidutils
from sqlalchemy import sql

from neutron_lib import context as ctx

from neutron.db import api as db_api
from neutron.db.models import portbinding
from neutron.db.models import segment
from neutron.db import models_v2
from neutron.objects import network as network_obj
from neutron.plugins.ml2 import models as ml2_models

from midonet.neutron.db import port_binding_db
from midonet.neutron.db import provider_network_db


# midonet v2 port binding:
#    midonet_port_bindings (port_binding_db.PortBindingInfo)
#    portbindingports (portbinding.PortBindingPort)

# ml2 port binding:
#    ml2_port_bindings (ml2_models.PortBinding)
#    ml2_port_binding_levels (ml2_models.PortBindingLevel)
#    ml2_distributed_port_bindings (ml2_models.DistributedPortBinding)

# midonet v2 segments:
#    midonet_network_bindings (provider_network_db.NetworkBinding)

# ml2 segments:
#    networksegments (neutron.db.models.segment.NetworkSegment)


LOG = logging.getLogger(__name__)


# NOTE(yamamoto): This module performs dirty things against ML2 tables,
# which we don't have any control of.  It's fragile and prone to be broken
# by future ML2 schema changes.  To avoid corrupting tables, we bail out
# on unknown neutron alembic versions.  This likely involves false positives
# but it's better safe than sorry.
_KNOWN_NEUTRON_ALEMBIC_VERSIONS = [
    '7d32f979895f',
    '5c85685d616d',
]


def log_calls(func):
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        LOG.info('Calling %(func)s with %(args)s %(kwargs)s', {
            'func': func.__name__,
            'args': args,
            'kwargs': kwargs,
        })
        return func(*args, **kwargs)

    return wrapper


def check_alembic_versions(context):
    LOG.info(
        'Checking Neutron alembic versions are in %s',
        _KNOWN_NEUTRON_ALEMBIC_VERSIONS)
    stmt = sql.select(
        [sql.column('version_num')]).select_from(sql.table('alembic_version'))
    for row in context.session.execute(stmt).fetchall():
        version = row[0]
        if version not in _KNOWN_NEUTRON_ALEMBIC_VERSIONS:
            LOG.error('Unknown version %(version)s', {
                'version': version,
            })
            raise SystemExit(1)
        LOG.info('Known version %(version)s', {
            'version': version,
        })


@log_calls
def add_segment(context, network_id, network_type):
    # NOTE(yamamoto): The code fragment is a modified copy of segments_db.py.
    # We don't want to make callback notifications.
    segment_id = uuidutils.generate_uuid()
    netseg_obj = network_obj.NetworkSegment(
        context,
        id=segment_id,
        network_id=network_id,
        network_type=network_type,
        physical_network=None,
        segmentation_id=None,
        segment_index=0,
        is_dynamic=False)
    netseg_obj.create()
    return segment_id


@log_calls
def add_binding_bound(context, port_id, segment_id, host, interface_name):
    context.session.add(ml2_models.PortBindingLevel(
        port_id=port_id,
        host=host,
        level=0,
        driver='midonet',
        segment_id=segment_id))
    profile = {}
    if interface_name is not None:
        profile['interface_name'] = interface_name
    context.session.add(ml2_models.PortBinding(
        port_id=port_id,
        host=host,
        vif_type='midonet',
        vnic_type='normal',
        profile=jsonutils.dumps(profile),
        vif_details=jsonutils.dumps({'port_filter': True}),
        status='ACTIVE'))


@log_calls
def add_binding_unbound(context, port_id):
    # ml2 add_port_binding equiv
    context.session.add(ml2_models.PortBinding(
        port_id=port_id,
        vif_type='unbound',
        vnic_type='normal',
        profile='',
        vif_details='',
        status='ACTIVE'))


@log_calls
def migrate():
    # Migrate db tables from v2 to ML2
    # NOTE(yamamoto): This would bump revisions of affected resources.

    context = ctx.get_admin_context()
    with db_api.context_manager.writer.using(context):
        # Lock all old rows for simplicity.  This doesn't benefit from
        # concurrency anyway.
        old_segments = context.session.query(
            provider_network_db.NetworkBinding).with_for_update().all()
        old_host_bindings = context.session.query(
            portbinding.PortBindingPort).with_for_update().all()
        old_interface_bindings = context.session.query(
            port_binding_db.PortBindingInfo).with_for_update().all()

        if not any([old_segments, old_host_bindings, old_interface_bindings]):
            LOG.info("Nothing to do; v2 plugin tables are empty")
            return

        if any([
            context.session.query(segment.NetworkSegment).count(),
            context.session.query(ml2_models.PortBinding).count(),
            context.session.query(ml2_models.PortBindingLevel).count(),
            context.session.query(ml2_models.DistributedPortBinding).count(),
        ]):
            LOG.error("Both of v2 plugin and ML2 tables have some contents")
            raise SystemExit(1)

        LOG.info("Moving data from v2 plugin tables to ML2 tables")
        check_alembic_versions(context)

        # Migrate network segments
        # NOTE(yamamoto): Only "uplink" networks have a NetworkBinding row.
        # A network without NetworkBinding row is of the default "midonet"
        # type.
        segments = {}
        uplink_network_ids = [seg.network_id for seg in old_segments]
        for network_id in uplink_network_ids:
            segments[network_id] = add_segment(
                context, network_id=network_id, network_type="uplink")
        networks = context.session.query(models_v2.Network).all()
        for net in networks:
            if net.id not in uplink_network_ids:
                segments[net.id] = add_segment(
                    context, network_id=net.id, network_type="midonet")

        # Migrate port bindings
        # NOTE(yamamoto): Unlike midonet v2, ML2 has PortBinding rows
        # even for unbound ports.
        port_host = {}
        for binding in old_host_bindings:
            port_host[binding.port_id] = binding.host
        port_interface = {}
        for binding in old_interface_bindings:
            port_interface[binding.port_id] = binding.interface_name
        for port in context.session.query(models_v2.Port).all():
            port_id = port.id
            if port_id in port_host:
                add_binding_bound(
                    context, port_id, segments[port.network_id],
                    port_host[port_id], port_interface.get(port_id))
            else:
                add_binding_unbound(context, port_id)

        # Delete no longer used rows
        map(context.session.delete, old_segments)
        map(context.session.delete, old_host_bindings)
        map(context.session.delete, old_interface_bindings)

    LOG.info("DB Migration from MidoNet v2 to ML2 completed successfully")



# Copyright 2015 Midokura SARL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""add agent membership

Revision ID: 19808c5df22a
Revises: d8a5c672761
Create Date: 2015-02-26 14:39:25.219125

"""

# revision identifiers, used by Alembic.
revision = '19808c5df22a'
down_revision = 'd8a5c672761'

from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        "midonet_agent_memberships",
        sa.Column('id', sa.String(36), primary_key=True),
        sa.Column('ip_address', sa.String(64), nullable=False),)



# Copyright (c) 2015 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as exc
from neutron_lib.plugins.ml2 import api
from oslo_log import log

from midonet.neutron.common import constants as const

LOG = log.getLogger(__name__)


class UplinkTypeDriver(api.ML2TypeDriver):
    """Type driver for Uplink networks

    This type driver differentiates uplinks networks from other types.
    """

    def __init__(self):
        LOG.info("ML2 UplinkTypeDriver initialization complete")

    def initialize(self):
        pass

    def get_type(self):
        return const.TYPE_UPLINK

    def is_partial_segment(self, segment):
        return False

    def validate_provider_segment(self, segment):
        for key, value in segment.items():
            if value and key != api.NETWORK_TYPE:
                msg = _("%s prohibited for uplink provider network") % key
                raise exc.InvalidInput(error_message=msg)

    def reserve_provider_segment(self, context, segment):
        return segment

    def allocate_tenant_segment(self, context):
        return {api.NETWORK_TYPE: const.TYPE_UPLINK}

    def release_segment(self, context, segment):
        pass

    def get_mtu(self, physical):
        pass



# Copyright (C) 2016 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import mock
from mox3 import mox

from neutronclient import shell
from neutronclient.tests.unit import test_cli20

TOKEN = test_cli20.TOKEN


class MyResp(test_cli20.MyResp):

    pass


class MyApp(test_cli20.MyApp):

    pass


class MyComparator(test_cli20.MyComparator):

    pass


class CLIExtTestV20Base(test_cli20.CLITestV20Base):

    def setUp(self, plurals=None):
        super(CLIExtTestV20Base, self).setUp(plurals=plurals)

    def _setup_mock_patch(self, name):
        patcher = mock.patch(name)
        thing = patcher.start()
        return thing

    def _mock_load_extensions(self, resource):
        load_method = ('neutronclient.common.extension.' +
                       '_discover_via_entry_points')
        load_ext_mock = self._setup_mock_patch(load_method)
        load_ext_mock.return_value = [resource]
        return load_ext_mock

    def _test_update_ext_resource(self, resource, cmd, myid, args,
                                  extrafields,
                                  cmd_resource=None, parent_id=None):
        self.mox.StubOutWithMock(cmd, "get_client")
        self.mox.StubOutWithMock(self.client.httpclient, "request")
        cmd.get_client().MultipleTimes().AndReturn(self.client)
        if not cmd_resource:
            cmd_resource = resource

        body = {resource: extrafields}
        path = getattr(self.client, cmd_resource + "_path")
        if parent_id:
            path = path % parent_id
        path = path % myid
        mox_body = MyComparator(body, self.client)

        self.client.httpclient.request(
            test_cli20.MyUrlComparator(
                test_cli20.end_url(path), self.client),
            'PUT',
            body=mox_body,
            headers=mox.ContainsKeyValue(
                'X-Auth-Token', TOKEN)).AndReturn((MyResp(204), None))
        self.mox.ReplayAll()
        cmd_parser = cmd.get_parser("update_" + cmd_resource)
        shell.run_command(cmd, cmd_parser, args)
        self.mox.VerifyAll()
        self.mox.UnsetStubs()
        _str = self.fake_stdout.make_string()
        self.assertIn(myid, _str)

    def _test_show_ext_resource(self, resource, cmd, myid, args, fields=(),
                                cmd_resource=None, parent_id=None):
        self.mox.StubOutWithMock(cmd, "get_client")
        self.mox.StubOutWithMock(self.client.httpclient, "request")
        cmd.get_client().MultipleTimes().AndReturn(self.client)
        if not cmd_resource:
            cmd_resource = resource

        query = "&".join(["fields=%s" % field for field in fields])
        expected_res = {resource:
                        {self.id_field: myid,
                         'name': 'myname', }, }
        resstr = self.client.serialize(expected_res)
        path = getattr(self.client, cmd_resource + "_path")
        if parent_id:
            path = path % parent_id
        path = path % myid
        self.client.httpclient.request(
            test_cli20.end_url(path, query), 'GET',
            body=None,
            headers=mox.ContainsKeyValue(
                'X-Auth-Token', TOKEN)).AndReturn((MyResp(200), resstr))
        self.mox.ReplayAll()
        cmd_parser = cmd.get_parser("show_" + cmd_resource)
        shell.run_command(cmd, cmd_parser, args)
        self.mox.VerifyAll()
        self.mox.UnsetStubs()
        _str = self.fake_stdout.make_string()
        self.assertIn(myid, _str)
        self.assertIn('myname', _str)

    def _test_delete_ext_resource(self, resource, cmd, myid, args,
                                  cmd_resource=None, parent_id=None):
        self.mox.StubOutWithMock(cmd, "get_client")
        self.mox.StubOutWithMock(self.client.httpclient, "request")
        cmd.get_client().MultipleTimes().AndReturn(self.client)
        if not cmd_resource:
            cmd_resource = resource
        path = getattr(self.client, cmd_resource + "_path")
        if parent_id:
            path = path % parent_id
        path = path % myid
        self.client.httpclient.request(
            test_cli20.end_url(path), 'DELETE',
            body=None,
            headers=mox.ContainsKeyValue(
                'X-Auth-Token', TOKEN)).AndReturn((MyResp(204), None))
        self.mox.ReplayAll()
        cmd_parser = cmd.get_parser("delete_" + cmd_resource)
        shell.run_command(cmd, cmd_parser, args)
        self.mox.VerifyAll()
        self.mox.UnsetStubs()
        _str = self.fake_stdout.make_string()
        self.assertIn(myid, _str)



# Copyright 2016 Midokura SARL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""add midonet gateway network vlan devices

Revision ID: 3df8cd6d2ada
Revises: cfe0dea89aa
Create Date: 2016-04-13 07:03:39.557871

"""

# revision identifiers, used by Alembic.
revision = '3df8cd6d2ada'
down_revision = 'cfe0dea89aa'

from alembic import op
import sqlalchemy as sa


def upgrade():
    op.create_table('midonet_gateway_network_vlan_devices',
                    sa.Column('device_id', sa.String(length=36),
                              nullable=False, primary_key=True),
                    sa.Column('resource_id', sa.String(length=36),
                              nullable=False),
                    sa.ForeignKeyConstraint(['device_id'],
                                            ['midonet_gateway_devices.id'],
                                            ondelete='CASCADE'),
                    sa.ForeignKeyConstraint(['resource_id'],
                                            ['networks.id']))



# Copyright 2016 Midokura SARL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""add foreign key constraint on gateway device

Revision ID: 7234c9979bf2
Revises: 143b26cc5196
Create Date: 2016-04-12 10:32:00.698251

"""

from alembic import op

from neutron.db import migration


# revision identifiers, used by Alembic.
revision = '7234c9979bf2'
down_revision = '143b26cc5196'
depends_on = ('cfe0dea89aa',)

# milestone identifier, used by neutron-db-manage
neutron_milestone = [migration.MITAKA]


def upgrade():
    op.create_foreign_key(
        constraint_name=None,
        source_table='midonet_gateway_overlay_router_devices',
        referent_table='routers',
        local_cols=['resource_id'],
        remote_cols=['id'])



# Copyright 2015 Midokura SARL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from neutron.db.migration.models import head

import midonet.neutron.db.bgp_speaker_router_insertion_model  # noqa
import midonet.neutron.db.data_state_db  # noqa
import midonet.neutron.db.data_version_db  # noqa
import midonet.neutron.db.gateway_device  # noqa
import midonet.neutron.db.logging_resource_model  # noqa
import midonet.neutron.db.port_binding_db  # noqa
import midonet.neutron.db.provider_network_db  # noqa
import midonet.neutron.db.task_db  # noqa


def get_metadata():
    return head.model_base.BASEV2.metadata



# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib

import mock
from oslo_utils import uuidutils
import webob.exc

from networking_l2gw.extensions import l2gateway
from networking_l2gw.extensions import l2gatewayconnection
from networking_l2gw.services.l2gateway.common import constants as l2gw_consts
from neutron.db import servicetype_db as st_db
from neutron.services import provider_configuration as provconf
from neutron.tests.unit.api import test_extensions as test_ex
from neutron.tests.unit.extensions import test_l3

from midonet.neutron.common import constants as mido_const
from midonet.neutron.tests.unit import test_extension_gateway_device as test_gw
from midonet.neutron.tests.unit import test_midonet_plugin_ml2 as test_mn_ml2

L2_GW_NAME = 'l2_gw1'
L2_GW_NAME2 = 'l2_gw2'
FAKE_SEG_ID = '1000'
FAKE_SEG_ID_VXLAN = '7000'
INVALID_VLAN_ID = 4095
INVALID_VXLAN_ID = 16777216
MN_PLUGIN_KLASS = 'midonet_l2gw'
MN_DRIVER_KLASS = ('midonet.neutron.services.l2gateway.service_drivers.'
                   'l2gw_midonet.MidonetL2gwDriver')


class MidonetL2GatewayTestExtensionManager(
        test_gw.GatewayDeviceTestExtensionManager,
        test_l3.L3TestExtensionManager):

    def get_resources(self):
        res = super(MidonetL2GatewayTestExtensionManager, self).get_resources()
        return (res + l2gateway.L2gateway.get_resources() +
                l2gatewayconnection.L2gatewayconnection.get_resources())

    def get_actions(self):
        return []

    def get_request_extensions(self):
        return []


class MidonetL2GatewayTestCaseMixin(test_gw.GatewayDeviceTestCaseMixin,
                                    test_l3.L3NatTestCaseMixin):

    def setUp(self, plugin=None, ext_mgr=None):

        service_plugins = {'l2gw_plugin_name': MN_PLUGIN_KLASS,
                           'gateway_device_plugin_name':
                               test_gw.DB_GATEWAY_DEVICE_PLUGIN_KLASS}
        l2gw_provider = (l2gw_consts.L2GW + ':' +
                         mido_const.MIDONET_L2GW_PROVIDER +
                         ':' + MN_DRIVER_KLASS + ':default')
        mock.patch.object(provconf.NeutronModule, 'service_providers',
                          return_value=[l2gw_provider]).start()
        manager = st_db.ServiceTypeManager.get_instance()
        manager.add_provider_configuration(
            l2gw_consts.L2GW, provconf.ProviderConfiguration())
        l2_gw_mgr = MidonetL2GatewayTestExtensionManager()

        super(MidonetL2GatewayTestCaseMixin, self).setUp(
            service_plugins=service_plugins, ext_mgr=l2_gw_mgr)
        self.ext_api = test_ex.setup_extensions_middleware(l2_gw_mgr)

        network = self._make_network(self.fmt, 'net1', True)
        self._network_id = network['network']['id']
        self._subnet = self._make_subnet(self.fmt, network, "10.0.0.1",
                                         '10.0.0.0/24')
        self._subnet_id = self._subnet['subnet']['id']
        network2 = self._make_network(self.fmt, 'net2', True)
        self._network_id2 = network2['network']['id']
        self._subnet2 = self._make_subnet(self.fmt, network2, "20.0.0.1",
                                          '20.0.0.0/24')
        self._subnet_id2 = self._subnet2['subnet']['id']
        router1 = self._make_router('json', uuidutils.generate_uuid(),
                                    'router1', True)
        self._router_id = router1['router']['id']
        router2 = self._make_router('json', uuidutils.generate_uuid(),
                                    'router2', True)
        self._router_id2 = router2['router']['id']

        # for network_vlan gateway device setting
        res = self._create_network(self.fmt, 'gateway_network_vlan', True)
        self._vlan_network_id = self.deserialize(
            self.fmt, res)['network']['id']

    def _create_l2_gateway(self, name=L2_GW_NAME, device_id="", device_id2="",
                           seg_id=None):
        data = {'l2_gateway': {'devices': [{'device_id': device_id}],
                               'tenant_id': uuidutils.generate_uuid(),
                               'name': name}}
        if device_id2:
            data['l2_gateway']['devices'].append({'device_id': device_id2})
        if seg_id:
            data['l2_gateway']['devices'][0]['segmentation_id'] = seg_id
        l2_gw_req = self.new_create_request('l2-gateways',
                                            data, self.fmt)
        return l2_gw_req.get_response(self.ext_api)

    def _create_l2_gateway_connection(self, l2_gateway_id="", network_id="",
                                      segmentation_id=FAKE_SEG_ID):
        data = {
            'l2_gateway_connection': {'l2_gateway_id': l2_gateway_id,
                                      'network_id': network_id,
                                      'tenant_id': uuidutils.generate_uuid(),
                                      'segmentation_id': segmentation_id}}
        l2_gw_conn_req = self.new_create_request('l2-gateway-connections',
                                                 data, self.fmt)
        return l2_gw_conn_req.get_response(self.ext_api)

    @contextlib.contextmanager
    def l2_gateway(self, name=L2_GW_NAME, device_id="", segmentation_id=''):
        l2_gw = self._make_l2_gateway(name, device_id, segmentation_id)
        yield l2_gw

    @contextlib.contextmanager
    def l2_gateway_connection(self, l2_gateway_id="", network_id="",
                              segmentation_id=FAKE_SEG_ID):
        l2_gw_con = self._make_l2_gateway_connection(l2_gateway_id,
                                                     network_id,
                                                     segmentation_id)
        yield l2_gw_con

    def _make_l2_gateway(self, name, device_id, segmentation_id):
        res = self._create_l2_gateway(name, device_id, seg_id=segmentation_id)
        if res.status_int >= webob.exc.HTTPBadRequest.code:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(self.fmt, res)

    def _make_l2_gateway_connection(self, l2_gateway_id, network_id,
                                    segmentation_id):
        res = self._create_l2_gateway_connection(l2_gateway_id, network_id,
                                                 segmentation_id)
        if res.status_int >= webob.exc.HTTPBadRequest.code:
            raise webob.exc.HTTPClientError(code=res.status_int)
        return self.deserialize(self.fmt, res)

    def test_create_midonet_l2gateway(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            name = L2_GW_NAME
            device_id = gw_dev['gateway_device']['id']
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']) as l2_gw:
                self.assertEqual(name, l2_gw['l2_gateway']['name'])
                self.assertEqual(
                    device_id,
                    l2_gw['l2_gateway']['devices'][0]['device_id'])

    def test_create_midonet_l2gateway_vlan(self):
        with self.gateway_device_type_network_vlan(
                resource_id=self._vlan_network_id) as gw_dev:
            name = L2_GW_NAME
            device_id = gw_dev['gateway_device']['id']
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']) as l2_gw:
                self.assertEqual(name, l2_gw['l2_gateway']['name'])
                self.assertEqual(
                    device_id,
                    l2_gw['l2_gateway']['devices'][0]['device_id'])

    def test_create_midonet_l2gateway_with_invalid_seg_id(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            res = self._create_l2_gateway(
                device_id=gw_dev['gateway_device']['id'],
                seg_id=INVALID_VXLAN_ID)
            self.deserialize(self.fmt, res)
            self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_create_midonet_l2gateway_vlan_with_invalid_seg_id(self):
        with self.gateway_device_type_network_vlan(
                resource_id=self._vlan_network_id) as gw_dev:
            res = self._create_l2_gateway(
                device_id=gw_dev['gateway_device']['id'],
                seg_id=INVALID_VLAN_ID)
            self.deserialize(self.fmt, res)
            self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_create_midonet_l2gateway_with_gateway_device_not_found(self):
        res = self._create_l2_gateway(device_id='a')
        self.deserialize(self.fmt, res)
        self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)

    def test_create_midonet_l2gateway_with_multiple_gateway_devices(self):
        res = self._create_l2_gateway(device_id='a', device_id2='b')
        self.deserialize(self.fmt, res)
        self.assertEqual(webob.exc.HTTPBadRequest.code, res.status_int)

    def test_create_midonet_l2gateway_with_segmentation_id(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            name = L2_GW_NAME
            device_id = gw_dev['gateway_device']['id']
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id'],
                    segmentation_id=FAKE_SEG_ID_VXLAN) as l2_gw:
                self.assertEqual(name, l2_gw['l2_gateway']['name'])
                self.assertEqual(
                    device_id,
                    l2_gw['l2_gateway']['devices'][0]['device_id'])
                self.assertEqual(
                    FAKE_SEG_ID_VXLAN,
                    l2_gw['l2_gateway']['devices'][0]['segmentation_id'])

    def test_create_midonet_l2gateway_vlan_with_segmentation_id(self):
        with self.gateway_device_type_network_vlan(
                resource_id=self._vlan_network_id) as gw_dev:
            name = L2_GW_NAME
            device_id = gw_dev['gateway_device']['id']
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id'],
                    segmentation_id=FAKE_SEG_ID) as l2_gw:
                self.assertEqual(name, l2_gw['l2_gateway']['name'])
                self.assertEqual(
                    device_id,
                    l2_gw['l2_gateway']['devices'][0]['device_id'])
                self.assertEqual(
                    FAKE_SEG_ID,
                    l2_gw['l2_gateway']['devices'][0]['segmentation_id'])

    def test_delete_midonet_l2gateway(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']) as l2_gw:
                req = self.new_delete_request('l2-gateways',
                                              l2_gw['l2_gateway']['id'])
                res = req.get_response(self.ext_api)
                self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_delete_midonet_l2gateway_vlan(self):
        with self.gateway_device_type_network_vlan(
                resource_id=self._vlan_network_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']) as l2_gw:
                req = self.new_delete_request('l2-gateways',
                                              l2_gw['l2_gateway']['id'])
                res = req.get_response(self.ext_api)
                self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_delete_midonet_l2gateway_with_seg_id(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id'],
                    segmentation_id=FAKE_SEG_ID) as l2_gw:
                req = self.new_delete_request('l2-gateways',
                                              l2_gw['l2_gateway']['id'])
                res = req.get_response(self.ext_api)
                self.assertEqual(webob.exc.HTTPNoContent.code, res.status_int)

    def test_list_l2gateways(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev1:
            with self.l2_gateway(name=L2_GW_NAME,
                                 device_id=gw_dev1['gateway_device']['id']):
                with self.gateway_device_type_router_vtep(
                        resource_id=self._router_id2,
                        tunnel_ips=[test_gw.FAKE_TUNNEL_IP2]) as gw_dev2:
                    with self.l2_gateway(
                            name=L2_GW_NAME,
                            device_id=gw_dev2['gateway_device']['id']):
                        req = self.new_list_request('l2-gateways')
                        res = self.deserialize(
                            self.fmt, req.get_response(self.ext_api))
                        self.assertEqual(2, len(res['l2_gateways']))

    def test_create_midonet_l2gateway_connection(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']) as l2_gw:
                with self.l2_gateway_connection(
                        l2_gateway_id=l2_gw['l2_gateway']['id'],
                        network_id=self._network_id,
                        segmentation_id=str(FAKE_SEG_ID_VXLAN)) as l2_gw_con:
                    self.assertEqual(
                        self._network_id,
                        l2_gw_con['l2_gateway_connection']['network_id'])
                    self.assertEqual(
                        l2_gw['l2_gateway']['id'],
                        l2_gw_con['l2_gateway_connection']['l2_gateway_id'])
                    self.assertEqual(
                        FAKE_SEG_ID_VXLAN,
                        l2_gw_con['l2_gateway_connection']['segmentation_id'])

    def test_create_midonet_l2gateway_vlan_connection(self):
        with self.gateway_device_type_network_vlan(
                resource_id=self._vlan_network_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']) as l2_gw:
                with self.l2_gateway_connection(
                        l2_gateway_id=l2_gw['l2_gateway']['id'],
                        network_id=self._network_id) as l2_gw_con:
                    self.assertEqual(
                        self._network_id,
                        l2_gw_con['l2_gateway_connection']['network_id'])
                    self.assertEqual(
                        l2_gw['l2_gateway']['id'],
                        l2_gw_con['l2_gateway_connection']['l2_gateway_id'])
                    self.assertEqual(
                        FAKE_SEG_ID,
                        l2_gw_con['l2_gateway_connection']['segmentation_id'])

    def test_create_midonet_l2gateway_and_l2gateway_con_without_seg_id(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']) as l2_gw:
                res = self._create_l2_gateway_connection(
                    l2_gateway_id=l2_gw['l2_gateway']['id'],
                    network_id=self._network_id,
                    segmentation_id='')
                self.deserialize(self.fmt, res)
                self.assertEqual(webob.exc.HTTPBadRequest.code,
                                 res.status_int)

    def test_create_midonet_l2gateway_connection_with_not_found_network(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']) as l2_gw:
                res = self._create_l2_gateway_connection(
                    l2_gateway_id=l2_gw['l2_gateway']['id'],
                    network_id='a')
                self.deserialize(self.fmt, res)
                self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)

    def test_create_midonet_l2gateway_connection_with_invalid_seg_id(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']) as l2_gw:
                res = self._create_l2_gateway_connection(
                    l2_gateway_id=l2_gw['l2_gateway']['id'],
                    network_id=self._network_id,
                    segmentation_id=str(INVALID_VXLAN_ID))
                self.deserialize(self.fmt, res)
                self.assertEqual(webob.exc.HTTPBadRequest.code,
                                 res.status_int)

    def test_create_midonet_l2gateway_vlan_conn_with_invalid_seg_id(self):
        with self.gateway_device_type_network_vlan(
                resource_id=self._vlan_network_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']) as l2_gw:
                res = self._create_l2_gateway_connection(
                    l2_gateway_id=l2_gw['l2_gateway']['id'],
                    network_id=self._network_id,
                    segmentation_id=str(INVALID_VLAN_ID))
                self.deserialize(self.fmt, res)
                self.assertEqual(webob.exc.HTTPBadRequest.code,
                                 res.status_int)

    def test_create_midonet_l2gateway_connection_with_invalid_l2_gateway(self):
        res = self._create_l2_gateway_connection(l2_gateway_id='a',
                                                 network_id=self._network_id,
                                                 segmentation_id=FAKE_SEG_ID)
        self.deserialize(self.fmt, res)
        self.assertEqual(webob.exc.HTTPNotFound.code, res.status_int)

    def test_create_midonet_l2gateway_con_seg_id_with_l2gw_seg_id(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id'],
                    segmentation_id=FAKE_SEG_ID) as l2_gw:
                res = self._create_l2_gateway_connection(
                    l2_gateway_id=l2_gw['l2_gateway']['id'],
                    network_id=self._network_id,
                    segmentation_id=str(FAKE_SEG_ID))
                self.deserialize(self.fmt, res)
                self.assertEqual(webob.exc.HTTPBadRequest.code,
                                 res.status_int)

    def test_create_midonet_l2gateway_connection_with_l2gw_seg_id(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id'],
                    segmentation_id=FAKE_SEG_ID_VXLAN) as l2_gw:
                with self.l2_gateway_connection(
                        l2_gateway_id=l2_gw['l2_gateway']['id'],
                        network_id=self._network_id,
                        segmentation_id='') as l2_gw_con:
                    self.assertEqual(
                        self._network_id,
                        l2_gw_con['l2_gateway_connection']['network_id'])
                    self.assertEqual(
                        l2_gw['l2_gateway']['id'],
                        l2_gw_con['l2_gateway_connection']['l2_gateway_id'])
                    self.assertEqual(
                        '',
                        l2_gw_con['l2_gateway_connection']['segmentation_id'])

    def test_create_midonet_l2gateway_vlan_connection_with_l2gw_seg_id(self):
        with self.gateway_device_type_network_vlan(
                resource_id=self._vlan_network_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id'],
                    segmentation_id=FAKE_SEG_ID) as l2_gw:
                with self.l2_gateway_connection(
                        l2_gateway_id=l2_gw['l2_gateway']['id'],
                        network_id=self._network_id,
                        segmentation_id='') as l2_gw_con:
                    self.assertEqual(
                        self._network_id,
                        l2_gw_con['l2_gateway_connection']['network_id'])
                    self.assertEqual(
                        l2_gw['l2_gateway']['id'],
                        l2_gw_con['l2_gateway_connection']['l2_gateway_id'])
                    self.assertEqual(
                        '',
                        l2_gw_con['l2_gateway_connection']['segmentation_id'])

    def test_create_midonet_l2gateway_conn_error_delete_neutron_resouce(self):
        self.client_mock.create_l2_gateway_connection.side_effect = Exception(
            "Fake Error")
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']) as l2_gw:
                try:
                    with self.l2_gateway_connection(
                            l2_gateway_id=l2_gw['l2_gateway']['id'],
                            network_id=self._network_id,
                            segmentation_id=FAKE_SEG_ID):
                        self.assertTrue(False)
                except webob.exc.HTTPClientError:
                    pass
                req = self.new_list_request('l2-gateway-connections')
                res = self.deserialize(self.fmt,
                                       req.get_response(self.ext_api))
                self.assertFalse(res['l2_gateway_connections'])

    def test_create_multiple_midonet_l2gateway_con_with_same_l2gateway(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']) as l2_gw:
                with self.l2_gateway_connection(
                        l2_gateway_id=l2_gw['l2_gateway']['id'],
                        network_id=self._network_id):
                    res = self._create_l2_gateway_connection(
                        l2_gateway_id=l2_gw['l2_gateway']['id'],
                        network_id=self._network_id2,
                        segmentation_id=FAKE_SEG_ID)
                    self.deserialize(self.fmt, res)
                    self.assertEqual(webob.exc.HTTPConflict.code,
                                     res.status_int)

    def test_delete_midonet_l2gateway_connection(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']) as l2_gw:
                with self.l2_gateway_connection(
                        l2_gateway_id=l2_gw['l2_gateway']['id'],
                        network_id=self._network_id) as l2_gw_con:
                    req = self.new_delete_request(
                        'l2-gateway-connections',
                        l2_gw_con['l2_gateway_connection']['id'])
                    res = req.get_response(self.ext_api)
                    self.assertEqual(webob.exc.HTTPNoContent.code,
                                     res.status_int)

    def test_delete_midonet_l2gateway_vlan_connection(self):
        with self.gateway_device_type_network_vlan(
                resource_id=self._vlan_network_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']) as l2_gw:
                with self.l2_gateway_connection(
                        l2_gateway_id=l2_gw['l2_gateway']['id'],
                        network_id=self._network_id) as l2_gw_con:
                    req = self.new_delete_request(
                        'l2-gateway-connections',
                        l2_gw_con['l2_gateway_connection']['id'])
                    res = req.get_response(self.ext_api)
                    self.assertEqual(webob.exc.HTTPNoContent.code,
                                     res.status_int)

    def test_delete_midonet_l2gateway_connection_without_seg_id(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id'],
                    segmentation_id=FAKE_SEG_ID) as l2_gw:
                with self.l2_gateway_connection(
                        l2_gateway_id=l2_gw['l2_gateway']['id'],
                        network_id=self._network_id,
                        segmentation_id='') as l2_gw_con:
                    req = self.new_delete_request(
                        'l2-gateway-connections',
                        l2_gw_con['l2_gateway_connection']['id'])
                    res = req.get_response(self.ext_api)
                    self.assertEqual(webob.exc.HTTPNoContent.code,
                                     res.status_int)

    def test_delete_midonet_l2gateway_with_connection(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']) as l2_gw:
                with self.l2_gateway_connection(
                        l2_gateway_id=l2_gw['l2_gateway']['id'],
                        network_id=self._network_id):
                    req = self.new_delete_request('l2-gateways',
                                                  l2_gw['l2_gateway']['id'])
                    res = req.get_response(self.ext_api)
                    self.assertEqual(webob.exc.HTTPConflict.code,
                                     res.status_int)

    def test_list_gateway_connections(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']) as l2_gw:
                with self.l2_gateway_connection(
                        l2_gateway_id=l2_gw['l2_gateway']['id'],
                        network_id=self._network_id):
                    req = self.new_list_request('l2-gateway-connections')
                    res = self.deserialize(
                        self.fmt, req.get_response(self.ext_api))
                    self.assertEqual(1, len(res['l2_gateway_connections']))

    def test_delete_gateway_device_in_use_by_l2gateway(self):
        with self.gateway_device_type_router_vtep(
                resource_id=self._router_id) as gw_dev:
            with self.l2_gateway(
                    name=L2_GW_NAME,
                    device_id=gw_dev['gateway_device']['id']):
                req = self.new_delete_request('gw/gateway_devices',
                                              gw_dev['gateway_device']['id'])
                res = req.get_response(self.ext_api)
                self.assertEqual(webob.exc.HTTPConflict.code,
                                 res.status_int)


class MidonetL2GatewayTestCaseML2(MidonetL2GatewayTestCaseMixin,
                                  test_mn_ml2.MidonetPluginML2TestCase):
    pass



# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import netaddr
from oslo_utils import excutils
from oslo_utils import uuidutils
from sqlalchemy import orm

from neutron_lib.api.definitions import l3 as l3_apidef
from neutron_lib.callbacks import events
from neutron_lib.callbacks import exceptions
from neutron_lib.callbacks import registry
from neutron_lib.callbacks import resources
from neutron_lib import constants as n_const
from neutron_lib import exceptions as n_exc
from neutron_lib.exceptions import l3 as l3_exc

from neutron.db import _resource_extend as resource_extend
from neutron.db import l3_gwmode_db
from neutron.db.models import l3 as l3_models
from neutron.db import models_v2
from neutron.objects import router as l3_obj
from neutron.plugins.common import utils as p_utils

from midonet.neutron._i18n import _

DEVICE_OWNER_FLOATINGIP = n_const.DEVICE_OWNER_FLOATINGIP


class MidonetL3DBMixin(l3_gwmode_db.L3_NAT_db_mixin):
    # TODO(kengo): This is temporary workaround until upstream adds a check
    # for router deletion in l3_db

    def _check_router_not_in_use(self, context, router_id):
        try:
            kwargs = {'context': context, 'router_id': router_id}
            registry.notify(
                resources.ROUTER, events.BEFORE_DELETE, self, **kwargs)
        except exceptions.CallbackFailure as e:
            with excutils.save_and_reraise_exception():
                if len(e.errors) == 1:
                    raise e.errors[0].error
                raise l3_exc.RouterInUse(router_id=router_id, reason=e)

    def get_router_for_floatingip(
            self, context, internal_port, internal_subnet,
            external_network_id):
        # REVISIT(yamamoto): These direct manipulation of core-plugin db
        # resources is not ideal.
        gw_port = orm.aliased(models_v2.Port, name="gw_port")
        routerport_qry = context.session.query(
            l3_models.RouterPort.router_id,
            models_v2.IPAllocation.ip_address
        ).join(
            models_v2.Port, models_v2.IPAllocation
        ).filter(
            models_v2.Port.network_id == internal_port['network_id'],
            l3_models.RouterPort.port_type.in_(
                n_const.ROUTER_INTERFACE_OWNERS
            ),
            models_v2.IPAllocation.subnet_id == internal_subnet['id']
        ).join(
            gw_port, gw_port.device_id == l3_models.RouterPort.router_id
        ).filter(
            gw_port.network_id == external_network_id,
        ).distinct()

        first_router_id = None
        for router_id, interface_ip in routerport_qry:
            if interface_ip == internal_subnet['gateway_ip']:
                return router_id
            if not first_router_id:
                first_router_id = router_id
        if first_router_id:
            return first_router_id

        raise l3_exc.ExternalGatewayForFloatingIPNotFound(
            subnet_id=internal_subnet['id'],
            external_network_id=external_network_id,
            port_id=internal_port['id'])

    def _subnet_has_fip(self, context, router_id, subnet_id):
        # Return True if the subnet has one of floating IPs for the router
        subnet = self._core_plugin.get_subnet(context, subnet_id)
        subnet_cidr = netaddr.IPNetwork(subnet['cidr'])
        fip_qry = context.session.query(l3_models.FloatingIP)
        fip_qry = fip_qry.filter_by(router_id=router_id)
        for fip_db in fip_qry:
            if netaddr.IPAddress(fip_db['floating_ip_address']) in subnet_cidr:
                return True
        return False

    def router_gw_port_has_floating_ips(self, context, router_id):
        router = self._get_router(context, router_id)
        return any([
            self._subnet_has_fip(context, router_id, ip['subnet_id'])
            for ip in router.gw_port['fixed_ips']
        ])

    def find_next_hop_for_fip(self, context, floatingip_db):
        # Find a next-hop address for a route from the floating_network_id
        # network to the floating-ip.
        # NOTE(tidwellr) use admin context here
        # tenant may not own the router and that's OK on a FIP association
        router_id = floatingip_db.router_id
        router = self._get_router(context.elevated(), router_id)
        gw_port = None
        for rp in router.attached_ports:
            if rp.port.network_id == floatingip_db.floating_network_id:
                gw_port = rp.port
                break
        if not gw_port:
            return None
        fip_addr = netaddr.IPAddress(floatingip_db.floating_ip_address)
        for fixed_ip in gw_port.fixed_ips:
            addr = netaddr.IPAddress(fixed_ip.ip_address)
            if addr.version == fip_addr.version:
                return fixed_ip.ip_address

    def _port_fixed_ips_for_floatingip(self, port):
        # Returns the fixed IP addresses on the given port preferring IPv4
        # over IPv6 ones
        port_ips = self._port_ipv4_fixed_ips(port)
        if not port_ips:
            port_ips = self._port_ipv6_fixed_ips(port)
        return port_ips

    def _port_ipv6_fixed_ips(self, port):
        return [ip for ip in port['fixed_ips']
                if netaddr.IPAddress(ip['ip_address']).version == 6]

    def _validate_network_for_floatingip(self, context, net_id):
        if not any(self._core_plugin._get_network(context, net_id).subnets):
            msg = _("Network %s does not contain any subnet") % net_id
            raise n_exc.BadRequest(resource='floatingip', msg=msg)

    # REVISIT(bikfalvi): This method is a copy of the base class method,
    # modified to use the _port_fixed_ips hook and replace the network
    # validation from IPv4 to any IP.
    # NOTE(yamamoto): And Floating IP QoS stuff commented out
    def _create_floatingip(self, context, floatingip,
                           initial_status=n_const.FLOATINGIP_STATUS_ACTIVE):
        fip = floatingip['floatingip']
        fip_id = uuidutils.generate_uuid()

        f_net_id = fip['floating_network_id']
        if not self._core_plugin._network_is_external(context, f_net_id):
            msg = _("Network %s is not a valid external network") % f_net_id
            raise n_exc.BadRequest(resource='floatingip', msg=msg)

        self._validate_network_for_floatingip(context, f_net_id)

        # This external port is never exposed to the tenant.
        # it is used purely for internal system and admin use when
        # managing floating IPs.

        port = {'tenant_id': '',  # tenant intentionally not set
                'network_id': f_net_id,
                'admin_state_up': True,
                'device_id': 'PENDING',
                'device_owner': DEVICE_OWNER_FLOATINGIP,
                'status': n_const.PORT_STATUS_NOTAPPLICABLE,
                'name': ''}
        # Both subnet_id and floating_ip_address are accepted, if
        # floating_ip_address is not in the subnet,
        # InvalidIpForSubnet exception will be raised.
        fixed_ip = {}
        if fip['subnet_id']:
            fixed_ip['subnet_id'] = fip['subnet_id']
        if fip['floating_ip_address']:
            fixed_ip['ip_address'] = fip['floating_ip_address']
        if fixed_ip:
            port['fixed_ips'] = [fixed_ip]

        # 'status' in port dict could not be updated by default, use
        # check_allow_post to stop the verification of system
        external_port = p_utils.create_port(self._core_plugin,
                                            context.elevated(),
                                            {'port': port},
                                            check_allow_post=False)

        with p_utils.delete_port_on_error(self._core_plugin,
                                          context.elevated(),
                                          external_port['id']),\
                context.session.begin(subtransactions=True):
            external_ips = self._port_fixed_ips_for_floatingip(external_port)
            if not external_ips:
                raise n_exc.ExternalIpAddressExhausted(net_id=f_net_id)

            floating_fixed_ip = external_ips[0]
            floating_ip_address = floating_fixed_ip['ip_address']
            floatingip_obj = l3_obj.FloatingIP(
                context,
                id=fip_id,
                project_id=fip['tenant_id'],
                status=initial_status,
                floating_network_id=fip['floating_network_id'],
                floating_ip_address=floating_ip_address,
                floating_port_id=external_port['id'],
                description=fip.get('description'))
            # Update association with internal port
            # and define external IP address
            assoc_result = self._update_fip_assoc(
                context, fip, floatingip_obj, external_port)
            floatingip_obj.create()
            floatingip_dict = self._make_floatingip_dict(
                floatingip_obj, process_extensions=False)
            if self._is_dns_integration_supported:
                dns_data = self._process_dns_floatingip_create_precommit(
                    context, floatingip_dict, fip)
            # NOTE(yamamoto): MidoNet doesn't have Floating IP QoS
            # if self._is_fip_qos_supported:
            #     self._process_extra_fip_qos_create(context, fip_id, fip)
            floatingip_obj = l3_obj.FloatingIP.get_object(
                context, id=floatingip_obj.id)
            floatingip_db = floatingip_obj.db_obj

            registry.notify(resources.FLOATING_IP, events.PRECOMMIT_CREATE,
                            self, context=context, floatingip=fip,
                            floatingip_id=fip_id,
                            floatingip_db=floatingip_db)

        self._core_plugin.update_port(context.elevated(), external_port['id'],
                                      {'port': {'device_id': fip_id}})
        registry.notify(resources.FLOATING_IP,
                        events.AFTER_UPDATE,
                        self._update_fip_assoc,
                        **assoc_result)

        if self._is_dns_integration_supported:
            self._process_dns_floatingip_create_postcommit(context,
                                                           floatingip_dict,
                                                           dns_data)
        # TODO(lujinluo): Change floatingip_db to floatingip_obj once all
        # codes are migrated to use Floating IP OVO object.
        resource_extend.apply_funcs(l3_apidef.FLOATINGIPS, floatingip_dict,
                                    floatingip_db)
        return floatingip_dict



# Copyright 2015 Midokura SARL
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

from neutron_lib.db import model_base
from oslo_serialization import jsonutils
from oslo_utils import uuidutils
import sqlalchemy as sa

import midonet.neutron.db.data_state_db as ds_db

CONF_ID = '00000000-0000-0000-0000-000000000001'

CREATE = "CREATE"
DELETE = "DELETE"
UPDATE = "UPDATE"
FLUSH = "FLUSH"

NETWORK = "NETWORK"
SUBNET = "SUBNET"
ROUTER = "ROUTER"
PORT = "PORT"
FLOATING_IP = "FLOATINGIP"
SECURITY_GROUP = "SECURITYGROUP"
SECURITY_GROUP_RULE = "SECURITYGROUPRULE"
POOL = "POOL"
VIP = "VIP"
HEALTH_MONITOR = "HEALTHMONITOR"
MEMBER = "MEMBER"
PORT_BINDING = "PORTBINDING"
CONFIG = "CONFIG"


OP_IMPORT = 'IMPORT'
OP_FLUSH = 'FLUSH'


TASK_STATE_TABLE = 'midonet_task_state'


class Task(model_base.BASEV2, model_base.HasProjectNoIndex):
    __tablename__ = 'midonet_tasks'

    id = sa.Column(sa.Integer(), primary_key=True)
    type = sa.Column(sa.String(length=36))
    data_type = sa.Column(sa.String(length=36))
    data = sa.Column(sa.Text())
    resource_id = sa.Column(sa.String(36))
    transaction_id = sa.Column(sa.String(40), nullable=False)
    created_at = sa.Column(sa.DateTime(), default=datetime.datetime.utcnow,
                           nullable=False)


def get_current_task_data(session):
    data = dict()
    for task in session.query(Task):
        if task.data_type not in data:
            data[task.data_type] = dict()
        if task.type == DELETE:
            data[task.data_type].pop(task.resource_id, None)
        else:
            data[task.data_type][task.resource_id] = task.data
    return data


def get_task_list(session, show_unprocessed):
    tasks = session.query(Task)
    if show_unprocessed:
        task_state = session.query(ds_db.DataState).one()
        lp_id = task_state.last_processed_id
        if lp_id is not None:
            tasks = tasks.filter(Task.id > lp_id)
    return tasks


def task_clean(session):
    task_state = session.query(ds_db.DataState).one()
    lp_id = task_state.last_processed_id
    task_state.update({'last_processed_id': None,
                       'updated_at': datetime.datetime.utcnow()})
    session.query(Task).filter(Task.id <= lp_id).delete()
    session.commit()


def create_task(context, type, task_id=None, data_type=None,
                resource_id=None, data=None):

    with context.session.begin(subtransactions=True):
        db = Task(id=task_id,
                  type=type,
                  tenant_id=context.tenant,
                  data_type=data_type,
                  data=None if data is None else jsonutils.dumps(data),
                  resource_id=resource_id,
                  transaction_id=context.request_id)
        context.session.add(db)


def create_config_task(session, data):
    data['id'] = CONF_ID
    with session.begin(subtransactions=True):
        db = Task(type=CREATE,
                  tenant_id=None,
                  data_type=CONFIG,
                  data=jsonutils.dumps(data),
                  resource_id=data['id'],
                  transaction_id=uuidutils.generate_uuid())
        session.add(db)


def create_port_binding_task(context, port_id, interface_name, host):
    data = {
        'id': port_id,
        'host_id': host,
        'interface_name': interface_name,
        'port_id': port_id
    }
    create_task(context, CREATE, data_type=PORT_BINDING, resource_id=port_id,
                data=data)


def delete_port_binding_task(context, port_id):
    create_task(context, DELETE, data_type=PORT_BINDING, resource_id=port_id)



# Copyright 2015 Midokura SARL
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""add dynamic routing

Revision ID: 421564f630b1
Revises: 3fe2bca71c72
Create Date: 2015-03-31 04:40:24.533580

"""

# revision identifiers, used by Alembic.
revision = '421564f630b1'
down_revision = '3fe2bca71c72'

from alembic import op
import sqlalchemy as sa


def upgrade():

    op.create_table(
        'midonet_routing_instances',
        sa.Column('id', sa.String(length=36), primary_key=True),
        sa.Column('router_id', sa.String(length=36), nullable=False),
        sa.Column('local_as', sa.Integer(), nullable=False),
        sa.Column('protocol', sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(['router_id'], ['routers.id']))

    op.create_table(
        'midonet_routing_peers',
        sa.Column('id', sa.String(length=36), primary_key=True),
        sa.Column('routing_instance_id', sa.String(length=36),
                  nullable=False),
        sa.Column('port_id', sa.String(length=36), nullable=False),
        sa.Column('peer_as', sa.Integer(), nullable=False),
        sa.Column('peer_address', sa.String(length=255), nullable=False),
        sa.ForeignKeyConstraint(['routing_instance_id'],
                                ['midonet_routing_instances.id']),
        sa.ForeignKeyConstraint(['port_id'], ['ports.id']))

    op.create_table(
        'midonet_advertise_route',
        sa.Column('id', sa.String(length=36), primary_key=True),
        sa.Column('routing_instance_id', sa.String(length=36),
                  nullable=False),
        sa.Column('destination', sa.String(length=36),
                  nullable=False),
        sa.ForeignKeyConstraint(['routing_instance_id'],
                                ['midonet_routing_instances.id']))



# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib.plugins import directory

from networking_l2gw.db.l2gateway import l2gateway_db
from networking_l2gw.db.l2gateway import l2gateway_models as models
from networking_l2gw.services.l2gateway.common import constants
from networking_l2gw.services.l2gateway import exceptions as l2gw_exc
from neutron.api import extensions as neutron_extensions

from midonet.neutron.common import constants as midonet_const
from midonet.neutron.services.l2gateway.common import l2gw_midonet_validators
from midonet.neutron.services.l2gateway import exceptions


class MidonetL2GatewayMixin(l2gateway_db.L2GatewayMixin):
    # Override L2GatewayMixin to customize for Midonet L2GW

    def _check_and_get_gw_dev_service(self):
        gw_plugin = directory.get_plugin(midonet_const.GATEWAY_DEVICE)
        if not gw_plugin:
            raise exceptions.MidonetL2GatewayUnavailable()
        return gw_plugin

    def _validate_any_seg_id_empty_in_interface_dict(self, devices):
        # HACK: Override this since this validation method is not
        # applicable in MidoNet.
        pass

    def validate_l2_gateway_connection_for_create(self, context,
                                                  l2_gateway_connection):
        # HACK: Override this since segmentation id validation is not
        # applicable in Midonet. After deactivating segmentation id validation
        # of l2gw_validators.validate_network_mapping_list in plugin side,
        # validate it in this method.
        super(MidonetL2GatewayMixin,
              self).validate_l2_gateway_connection_for_create(
                  context, l2_gateway_connection)

        # Validate l2 gateway existence before getting gateway device type
        gw_connection = l2_gateway_connection[self.connection_resource]
        l2gw = self.get_l2_gateway(context, gw_connection['l2_gateway_id'])
        if not l2gw:
            raise l2gw_exc.L2GatewayNotFound(
                gateway_id=gw_connection['l2_gateway_id'])
        if self._get_l2_gateway_connection_by_l2gw_id(
                context, gw_connection['l2_gateway_id']):
            raise exceptions.MidonetL2GatewayConnectionExists(
                l2_gateway_id=gw_connection['l2_gateway_id'])

        # Validate segmentation id range according to gateway device type
        gw_connection = l2_gateway_connection[
            constants.CONNECTION_RESOURCE_NAME]
        seg_id = gw_connection.get(constants.SEG_ID)
        if seg_id:
            gw_type = self.get_gateway_device_type_from_l2gw(context, l2gw)
            l2gw_midonet_validators.is_valid_segmentaion_id(gw_type, seg_id)

    def _get_l2_gateway_seg_id(self, context, l2_gw_id):
        seg_id = None
        l2_gw_dev = self.get_l2gateway_devices_by_gateway_id(
            context, l2_gw_id)
        interfaces = self.get_l2gateway_interfaces_by_device_id(
            context, l2_gw_dev[0]['id'])
        if interfaces:
            seg_id = interfaces[0][constants.SEG_ID]
        return seg_id

    def _get_l2gw_devices_by_device_id(self, context, device_id):
        return context.session.query(models.L2GatewayDevice).filter_by(
            device_name=device_id).all()

    def get_gateway_device_type_from_l2gw(self, context, l2gw):
        gw_id = (l2gw['devices'][0].get('device_id')) or (
            l2gw['devices'][0].get('device_name'))
        gw_db = (directory.get_plugin(midonet_const.GATEWAY_DEVICE).
                 get_gateway_device(context, gw_id))
        return gw_db['type']

    def create_l2_gateway(self, context, l2_gateway):
        # HACK: set the device_name to device_id so that the networking-l2gw
        # DB class does not throw an error.
        gw = l2_gateway[self.gateway_resource]

        gw_plugin = self._check_and_get_gw_dev_service()
        for device in gw['devices']:
            gw = gw_plugin.get_gateway_device(context, device['device_id'])
            device['device_name'] = device['device_id']
            if device.get(constants.SEG_ID):
                l2gw_midonet_validators.is_valid_segmentaion_id(
                    gw['type'], device[constants.SEG_ID])
                device['interfaces'].append(
                    {constants.SEG_ID: [str(device[constants.SEG_ID])]})
        return super(MidonetL2GatewayMixin, self).create_l2_gateway(
            context, l2_gateway)

    def _make_l2_gateway_dict(self, l2_gateway, fields=None):
        l2gw = super(MidonetL2GatewayMixin, self)._make_l2_gateway_dict(
            l2_gateway, fields=fields)

        # HACK: change the 'device_name' to 'device_id' to match the API that
        # Midonet L2GW expects
        if 'devices' in l2gw:
            for device in l2gw['devices']:
                device['device_id'] = device['device_name']
                if device['interfaces']:
                    interface = device['interfaces'][0]
                    device[constants.SEG_ID] = interface[constants.SEG_ID][0]
                del device['device_name']
                del device['id']
                del device['interfaces']
        return l2gw

    def create_l2_gateway_connection(self, context, l2_gateway_connection):
        gw_connection = l2_gateway_connection[self.connection_resource]

        # Validate only network existence since l2_gateway existence is
        # validated in validate_l2_gateway_connection_for_create method.
        if not self._core_plugin.get_network(
                context, gw_connection['network_id']):
            raise neutron_extensions.NetworkNotFound(
                net_id=gw_connection['network_id'])
        return super(MidonetL2GatewayMixin, self).create_l2_gateway_connection(
            context, l2_gateway_connection)

    def update_l2_gateway(self, context, id, l2_gateway):
        raise NotImplementedError()



# Copyright (C) 2015 Midokura SARL.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sqlalchemy as sa

from neutron_lib.db import model_base

from midonet.neutron.common import exceptions as exc


DATA_STATE_TABLE = 'midonet_data_state'


class DataState(model_base.BASEV2):
    __tablename__ = DATA_STATE_TABLE
    id = sa.Column(sa.Integer(), primary_key=True)
    last_processed_task_id = sa.Column(sa.Integer(),
                                       sa.ForeignKey('midonet_tasks.id'))
    updated_at = sa.Column(sa.DateTime(), nullable=False)
    active_version = sa.Column(sa.Integer(),
                               sa.ForeignKey('midonet_data_versions.id'))
    readonly = sa.Column(sa.Boolean(), nullable=False)


def get_data_state(session):
    try:
        return session.query(DataState).one()
    except sa.orm.exc.NoResultFound:
        issue = "Missing Data State table"
        raise exc.InvalidMidonetDataState(issue)
    except sa.orm.exc.MultipleResultsFound:
        issue = "There should be exactly one Data State table"
        raise exc.InvalidMidonetDataState(issue)


def set_data_state_readonly(session, val):
    session.query(DataState).update({'readonly': val})
    session.commit()


def set_readonly(session):
    set_data_state_readonly(session, True)


def set_readwrite(session):
    set_data_state_readonly(session, False)



# Copyright (C) 2016 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import sys

from neutronclient import shell

from midonet.neutron.tests.unit.neutronclient_ext import test_cli20
from midonet.neutronclient.gateway_device_extension import _remote_mac_entry


class CLITestV20RemoteMacEntryJSON(test_cli20.CLIExtTestV20Base):

    def setUp(self):
        remote_mac_entry = ("remote_mac_entry", _remote_mac_entry)
        self._mock_load_extensions(remote_mac_entry)
        super(CLITestV20RemoteMacEntryJSON,
              self).setUp(plurals={'remote_mac_entries': 'remote_mac_entry'})
        self.register_non_admin_status_resource('remote_mac_entry')

    def test_remote_mac_entry_cmd_loaded(self):
        neutron_shell = shell.NeutronShell('2.0')
        remote_mac_entry_cmd = {'gateway-device-remote-mac-entry-list':
                                _remote_mac_entry.RemoteMacEntryList,
                                'gateway-device-remote-mac-entry-create':
                                _remote_mac_entry.RemoteMacEntryCreate,
                                'gateway-device-remote-mac-entry-delete':
                                _remote_mac_entry.RemoteMacEntryDelete,
                                'gateway-device-remote-mac-entry-show':
                                _remote_mac_entry.RemoteMacEntryShow}
        for cmd_name, cmd_class in remote_mac_entry_cmd.items():
            found = neutron_shell.command_manager.find_command([cmd_name])
            self.assertEqual(cmd_class, found[0])

    def _create_remote_mac_entry(self, args, position_names,
                                 position_values, parent_id=None):
        resource = 'remote_mac_entry'
        cmd = _remote_mac_entry.RemoteMacEntryCreate(
            test_cli20.MyApp(sys.stdout), None)
        self._test_create_resource(resource, cmd, None, 'myid',
                                   args, position_names, position_values,
                                   parent_id=parent_id)

    def test_create_remote_mac_entry(self):
        gw_device_id = 'my_gw_device'
        mac_addr = 'fa:16:3e:db:79:80'
        vtep_addr = '192.168.100.1'
        seg_id = '200'
        args = ['--mac-address', mac_addr, '--vtep-address', vtep_addr,
                '--segmentation-id', seg_id, gw_device_id]
        position_names = ['mac_address', 'vtep_address', 'segmentation_id']
        position_values = [mac_addr, vtep_addr, seg_id]
        self._create_remote_mac_entry(args, position_names,
                                      position_values, parent_id=gw_device_id)

    def test_create_remote_mac_entry_with_missing_gw_device_id(self):
        mac_addr = ''
        vtep_addr = ''
        seg_id = '200'
        args = ['--mac-address', mac_addr,
                '--vtep-address', vtep_addr, '--segmentation-id', seg_id]
        position_names = []
        position_values = []
        self.assertRaises(SystemExit, self._create_remote_mac_entry,
                          args, position_names, position_values)

    def test_create_remote_mac_entry_with_missing_seg_id(self):
        gw_device_id = 'my_gw_device'
        mac_addr = ''
        vtep_addr = ''
        args = ['--mac-address', mac_addr, '--vtep-address', vtep_addr,
                gw_device_id]
        position_names = []
        position_values = []
        self.assertRaises(SystemExit, self._create_remote_mac_entry,
                          args, position_names, position_values,
                          parent_id=gw_device_id)

    def test_delete_remote_mac_entry(self):
        resource = 'remote_mac_entry'
        cmd = _remote_mac_entry.RemoteMacEntryDelete(
            test_cli20.MyApp(sys.stdout), None)
        gw_device_id = 'my_gw_device'
        my_id = 'myid'
        args = [my_id, gw_device_id]
        self._test_delete_ext_resource(resource, cmd, my_id, args,
                                       parent_id=gw_device_id)

    def test_list_remote_mac_entries(self):
        resources = 'remote_mac_entries'
        cmd = _remote_mac_entry.RemoteMacEntryList(
            test_cli20.MyApp(sys.stdout), None)
        gw_device_id = 'my_gw_device'
        args = [gw_device_id]
        self._test_list_resources(resources, cmd, base_args=args,
                                  parent_id=gw_device_id)

    def test_show_remote_mac_entry(self):
        resource = 'remote_mac_entry'
        cmd = _remote_mac_entry.RemoteMacEntryShow(
            test_cli20.MyApp(sys.stdout), None)
        gw_device_id = 'my_gw_device'
        my_id = 'myid'
        args = [my_id, gw_device_id]
        self._test_show_ext_resource(resource, cmd, my_id, args,
                                     parent_id=gw_device_id)



# Copyright (c) 2015 Midokura SARL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron_lib import exceptions as exc
from neutron_lib.plugins.ml2 import api

from neutron.tests import base

from midonet.neutron.common import constants as const
from midonet.neutron.ml2 import type_midonet


class MidonetTypeTest(base.BaseTestCase):

    network_type = const.TYPE_MIDONET
    driver = type_midonet.MidonetTypeDriver()

    def setUp(self):
        super(MidonetTypeTest, self).setUp()
        self.context = None

    def test_is_partial_segment(self):
        segment = {api.NETWORK_TYPE: self.network_type}
        self.assertFalse(self.driver.is_partial_segment(segment))

    def test_validate_provider_segment(self):
        segment = {api.NETWORK_TYPE: self.network_type}
        self.driver.validate_provider_segment(segment)

    def test_validate_provider_segment_with_unallowed_physical_network(self):
        segment = {api.NETWORK_TYPE: self.network_type,
                   api.PHYSICAL_NETWORK: 'phys_net'}
        self.assertRaises(exc.InvalidInput,
                          self.driver.validate_provider_segment,
                          segment)

    def test_validate_provider_segment_with_unallowed_segmentation_id(self):
        segment = {api.NETWORK_TYPE: self.network_type,
                   api.SEGMENTATION_ID: 2}
        self.assertRaises(exc.InvalidInput,
                          self.driver.validate_provider_segment,
                          segment)

    def test_reserve_provider_segment(self):
        segment = {api.NETWORK_TYPE: self.network_type}
        observed = self.driver.reserve_provider_segment(self.context, segment)
        self.assertEqual(segment, observed)

    def test_release_provider_segment(self):
        segment = {api.NETWORK_TYPE: self.network_type}
        observed = self.driver.reserve_provider_segment(self.context, segment)
        self.driver.release_segment(self.context, observed)

    def test_allocate_tenant_segment(self):
        expected = {api.NETWORK_TYPE: self.network_type}
        observed = self.driver.allocate_tenant_segment(self.context)
        self.assertEqual(expected, observed)
